import os
import django
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "proj.settings")
django.setup()

import dataclasses
from datetime import date
from django.db.models import Count, Q
from app.models import Staff

today = date(2023, 2, 14)  # 動作検証用

# データクラスの定義
@dataclasses.dataclass
class DeliveryDesc:
    pk: int                        # 配達者pk
    name: str                      # 配達者名
    date: date                     # 配達日
    delivery_num: int|None = None  # 配達数
    unknown_num: int|None = None   # 宛先不明数

print("### 元のクエリ")

# ORMクエリの実装
qs = (
    Staff.objects
    # .only("pk", "name")  # group byのキー
    .annotate(
        delivery_num=Count("delivery", filter=Q(
            delivery__date=today,
            delivery__receiver__isnull=False,
        )),
        unknown_num=Count("delivery", filter=Q(
            delivery__date=today,
            delivery__receiver__isnull=True,
        )),
    )
)

# テンプレートへ渡すデータの詰め替え
desc_list = []
for staff in qs:
    desc = DeliveryDesc(
        pk=staff.pk,
        name=staff.name,
        date=today,
        delivery_num=staff.delivery_num,
        unknown_num=staff.unknown_num,
    )
    desc_list.append(desc)

# テンプレートでの参照
for desc in desc_list:
    print(desc)

# SQLのGROUP BY にある staff.last_login は .only("id", "name") で消せます。
# 1節目を受けて「.onlyを使わない」という方針になったコードという扱いです。

# ----------------------------

print("### 改善版")
# ORMクエリの実装
qs = (
    Staff.objects
    # .only("pk", "name")  # annotateがあるとonlyはgroup byに影響しない
    # .values("pk", "name")  # group byのキー
    .annotate(
        delivery_num=Count("delivery", filter=Q(
            delivery__date=today,
            delivery__receiver__isnull=False,
        )),
        unknown_num=Count("delivery", filter=Q(
            delivery__date=today,
            delivery__receiver__isnull=True,
        )),
    )
).values("pk", "name", "delivery_num", "unknown_num")

# クエリーセットを辞書のリストに展開。データの詰め替えは不要
values = list(qs)

# テンプレートでの参照
print(today)  # todayは各行に持たせず別で渡せばよい
for staff in values:
    print(staff)

# GROUP BY にある staff.last_login は .values("pk", "name") で消せます。
# 消さなくても害はありません。

# ----------------------------

print("### printsqlによるSQL出力")

def printsql(query):
    from sqlparse import format as sfmt
    print(sfmt(str(query), reindent_aligned=True))

printsql(qs.query)

# ----------------------------

print("### 理想のSQLからORMを組んだ改善版")
from django.db.models import FilteredRelation, F
# ORMクエリの実装
qs = (
    Staff.objects
    .annotate(
        dlist=FilteredRelation(
            "delivery",
            condition=Q(delivery__date=today)
        ),
        delivery_num=Count("dlist__receiver"),
        unknown_num=Count(
            "dlist",
            filter=Q(dlist__receiver__isnull=True),
        )
    ).values("pk", "name", "delivery_num", "unknown_num")
)

printsql(qs.query)
